#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <arpa/inet.h>
#include <curl/curl.h>
#include <sys/socket.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/epoll.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>
#include <vector>
#include <string>
#include <map>
#include <set>

#include "metacube.h"
#include "parse.h"
#include "server.h"
#include "serverpool.h"
#include "input.h"
#include "state.pb.h"

#define STREAM_ID "stream"
#define STREAM_URL "http://gruessi.zrh.sesse.net:4013/"

using namespace std;

ServerPool *servers = NULL;
volatile bool hupped = false;

void hup(int ignored)
{
	hupped = true;
}

int create_server_socket(int port)
{
	int server_sock = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
	if (server_sock == -1) {
		perror("socket");
		exit(1);
	}

	int one = 1;
	if (setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
		perror("setsockopt(SO_REUSEADDR)");
		exit(1);
	}

	// We want dual-stack sockets. (Sorry, OpenBSD and Windows XP...)
	int zero = 0;
	if (setsockopt(server_sock, IPPROTO_IPV6, IPV6_V6ONLY, &zero, sizeof(zero)) == -1) {
		perror("setsockopt(IPV6_V6ONLY)");
		exit(1);
	}

	sockaddr_in6 addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin6_family = AF_INET6;
	addr.sin6_port = htons(port);

	if (bind(server_sock, reinterpret_cast<sockaddr *>(&addr), sizeof(addr)) == -1) {
		perror("bind");
		exit(1);
	}

	if (listen(server_sock, 128) == -1) {
		perror("listen");
		exit(1);
	}

	return server_sock;
}

void *acceptor_thread_run(void *arg)
{
	int server_sock = int(intptr_t(arg));
	for ( ;; ) {
		sockaddr_in6 addr;
		socklen_t addrlen = sizeof(addr);

		// Get a new socket.
		int sock = accept(server_sock, reinterpret_cast<sockaddr *>(&addr), &addrlen);
		if (sock == -1 && errno == EINTR) {
			continue;
		}
		if (sock == -1) {
			perror("accept");
			exit(1);
		}

		// Set the socket as nonblocking.
		int one = 1;
		if (ioctl(sock, FIONBIO, &one) == -1) {
			perror("FIONBIO");
			exit(1);
		}

		// Pick a server, round-robin, and hand over the socket to it.
		servers->add_client(sock);
	}
}

// Serialize the given state to a file descriptor, and return the (still open)
// descriptor.
int make_tempfile(const CubemapStateProto &state)
{
	char tmpl[] = "/tmp/cubemapstate.XXXXXX";
	int state_fd = mkstemp(tmpl);
	if (state_fd == -1) {
		perror("mkstemp");
		exit(1);
	}

	string serialized;
	state.SerializeToString(&serialized);

	const char *ptr = serialized.data();
	size_t to_write = serialized.size();
	while (to_write > 0) {
		ssize_t ret = write(state_fd, ptr, to_write);
		if (ret == -1) {
			perror("write");
			exit(1);
		}

		ptr += ret;
		to_write -= ret;
	}

	return state_fd;
}

// Read the state back from the file descriptor made by make_tempfile,
// and close it.
CubemapStateProto read_tempfile(int state_fd)
{
	if (lseek(state_fd, 0, SEEK_SET) == -1) {
		perror("lseek");
		exit(1);
	}

	string serialized;
	char buf[4096];
	for ( ;; ) {
		ssize_t ret = read(state_fd, buf, sizeof(buf));
		if (ret == -1) {
			perror("read");
			exit(1);
		}
		if (ret == 0) {
			// EOF.
			break;
		}

		serialized.append(string(buf, buf + ret));
	}

	close(state_fd);  // Implicitly deletes the file.

	CubemapStateProto state;
	if (!state.ParseFromString(serialized)) {
		fprintf(stderr, "PANIC: Failed deserialization of state.\n");
		exit(1);
	}

	return state;
}

int main(int argc, char **argv)
{
	fprintf(stderr, "\nCubemap starting.\n");

	string config_filename = (argc == 1) ? "cubemap.config" : argv[1];
	vector<ConfigLine> config = parse_config(config_filename);

	int port = fetch_config_int(config, "port", 1, 65535);
	int num_servers = fetch_config_int(config, "num_servers", 1, 20000);  // Insanely high max limit.

	servers = new ServerPool(num_servers);

	int server_sock = -1, old_port = -1;
	set<string> deserialized_stream_ids;
	if (argc == 4 && strcmp(argv[2], "-state") == 0) {
		fprintf(stderr, "Deserializing state from previous process... ");
		int state_fd = atoi(argv[3]);
		CubemapStateProto loaded_state = read_tempfile(state_fd);

		// Deserialize the streams.
		for (int i = 0; i < loaded_state.streams_size(); ++i) {
			servers->add_stream_from_serialized(loaded_state.streams(i));
			deserialized_stream_ids.insert(loaded_state.streams(i).stream_id());
		}

		// Put back the existing clients. It doesn't matter which server we
		// allocate them to, so just do round-robin.
		for (int i = 0; i < loaded_state.clients_size(); ++i) {
			servers->add_client_from_serialized(loaded_state.clients(i));
		}

		// Deserialize the server socket.
		server_sock = loaded_state.server_sock();
		old_port = loaded_state.port();

		fprintf(stderr, "done.\n");
	}

	// Find all streams in the configuration file, and create them.
	set<string> expecting_stream_ids = deserialized_stream_ids;
	for (unsigned i = 0; i < config.size(); ++i) {
		if (config[i].keyword != "stream") {
			continue;
		}
		if (config[i].arguments.size() != 1) {
			fprintf(stderr, "ERROR: 'stream' takes exactly one argument\n");
			exit(1);
		}
		string stream_id = config[i].arguments[0];
		if (deserialized_stream_ids.count(stream_id) == 0) {
			servers->add_stream(stream_id);
		}
		expecting_stream_ids.erase(stream_id);
	}

	// Warn about any servers we've lost.
	// TODO: Make an option (delete=yes?) to actually shut down streams.
	for (set<string>::const_iterator stream_it = expecting_stream_ids.begin();
	     stream_it != expecting_stream_ids.end();
	     ++stream_it) {
		fprintf(stderr, "WARNING: stream '%s' disappeared from the configuration file.\n",
			stream_it->c_str());
		fprintf(stderr, "         It will not be deleted, but clients will not get any new inputs.\n");
	}

	// Open a new server socket if we do not already have one, or if we changed ports.
	if (server_sock != -1 && port != old_port) {
		fprintf(stderr, "NOTE: Port changed from %d to %d; opening new socket.\n", old_port, port);
		close(server_sock);
		server_sock = -1;
	}
	if (server_sock == -1) {
		server_sock = create_server_socket(port);
	}

	servers->run();

	pthread_t acceptor_thread;
	pthread_create(&acceptor_thread, NULL, acceptor_thread_run, reinterpret_cast<void *>(server_sock));

	// Find all streams in the configuration file, and create inputs for them.
	vector<Input *> inputs;
	for (unsigned i = 0; i < config.size(); ++i) {
		if (config[i].keyword != "stream") {
			continue;
		}
		assert(config[i].arguments.size() == 1);
		string stream_id = config[i].arguments[0];

		if (config[i].parameters.count("src") == 0) {
			fprintf(stderr, "WARNING: stream '%s' has no src= attribute, clients will not get any data.\n",
				stream_id.c_str());
			continue;
		}

		string src = config[i].parameters["src"];
		Input *input = new Input(stream_id, src);
		input->run();
		inputs.push_back(input);
	}

	signal(SIGHUP, hup);

	while (!hupped) {
		usleep(100000);
	}

	// OK, we've been HUPed. Time to shut down everything, serialize, and re-exec.
	for (size_t i = 0; i < inputs.size(); ++i) {
		inputs[i]->stop();
		delete inputs[i];  // TODO: serialize instead of using libcurl.
	}

	CubemapStateProto state;
	state.set_server_sock(server_sock);
	state.set_port(port);
	for (int i = 0; i < num_servers; ++i) {	
		servers->get_server(i)->stop();

		CubemapStateProto local_state = servers->get_server(i)->serialize();

		// The stream state should be identical between the servers, so we only store it once.
		if (i == 0) {
			state.mutable_streams()->MergeFrom(local_state.streams());
		}
		for (int j = 0; j < local_state.clients_size(); ++j) {
			state.add_clients()->MergeFrom(local_state.clients(j));
		}
	}
	delete servers;

	fprintf(stderr, "Serializing state and re-execing...\n");
	int state_fd = make_tempfile(state);

	char buf[16];
	sprintf(buf, "%d", state_fd);

	for ( ;; ) {
		execlp(argv[0], argv[0], config_filename.c_str(), "-state", buf, NULL);
		perror("execlp");
		fprintf(stderr, "PANIC: re-exec of %s failed. Waiting 0.2 seconds and trying again...\n", argv[0]);
		usleep(200000);
	}
}
