#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <vector>
#include <string>

#include "parse.h"

using namespace std;

vector<string> split_tokens(const string &line)
{
	vector<string> ret;
	string current_token;

	for (size_t i = 0; i < line.size(); ++i) {
		if (isspace(line[i])) {
			if (!current_token.empty()) {
				ret.push_back(current_token);
			}
			current_token.clear();
		} else {
			current_token.push_back(line[i]);
		}
	}
	if (!current_token.empty()) {
		ret.push_back(current_token);
	}
	return ret;
}

vector<string> split_lines(const string &str)
{
	vector<string> ret;
	string current_line;

	for (size_t i = 0; i < str.size(); ++i) {
		// Skip \r if followed by an \n.
		if (str[i] == '\r' && i < str.size() - 1 && str[i + 1] == '\n') {
			continue;
		}

		// End of the current line?
		if (str[i] == '\n') {
			if (!current_line.empty()) {
				ret.push_back(current_line);
			}
			current_line.clear();
		} else {
			current_line.push_back(str[i]);
		}
	}
	if (!current_line.empty()) {
		ret.push_back(current_line);
	}
	return ret;
}

vector<ConfigLine> parse_config(const string &filename)
{
	vector<ConfigLine> ret;

	FILE *fp = fopen(filename.c_str(), "r");
	if (fp == NULL) {
		perror(filename.c_str());
		exit(1);
	}

	char buf[4096];
	while (!feof(fp)) {
		if (fgets(buf, sizeof(buf), fp) == NULL) {
			break;
		}

		// Chop off the string at the first #, \r or \n.
		buf[strcspn(buf, "#\r\n")] = 0;

		// Remove all whitespace from the end of the string.
		size_t len = strlen(buf);
		while (len > 0 && isspace(buf[len - 1])) {
			buf[--len] = 0;
		}

		// If the line is now all blank, ignore it.
		if (len == 0) {
			continue;
		}

		vector<string> tokens = split_tokens(buf);
		assert(!tokens.empty());
		
		ConfigLine line;
		line.keyword = tokens[0];

		for (size_t i = 1; i < tokens.size(); ++i) {
			// foo=bar is a parameter; anything else is an argument.
			size_t equals_pos = tokens[i].find_first_of('=');
			if (equals_pos == string::npos) {
				line.arguments.push_back(tokens[i]);
			} else {
				string key = tokens[i].substr(0, equals_pos);
				string value = tokens[i].substr(equals_pos + 1, string::npos);
				line.parameters.insert(make_pair(key, value));
			}
		}

		ret.push_back(line);
	}

	fclose(fp);
	return ret;
}

int fetch_config_int(const vector<ConfigLine> &config, const string &keyword, int min_limit, int max_limit)
{
	bool value_found = false;
	int value = -1;
	for (unsigned i = 0; i < config.size(); ++i) {
		if (config[i].keyword != keyword) {
			continue;
		}
		if (config[i].parameters.size() > 0 ||
		    config[i].arguments.size() != 1) {
			fprintf(stderr, "ERROR: '%s' takes one argument and no parameters\n", keyword.c_str());
			exit(1);
		}
		value_found = true;
		value = atoi(config[i].arguments[0].c_str());  // TODO: verify int validity.
	}
	if (!value_found) {
		fprintf(stderr, "ERROR: Missing '%s' statement in config file.\n",
			keyword.c_str());
		exit(1);
	}
	if (value < min_limit || value > max_limit) {
		fprintf(stderr, "ERROR: '%s' is set to %d, must be in [%d,%d]\n",
			keyword.c_str(), value, min_limit, max_limit);
		exit(1);
	}
	return value;
}
